

!!$module sqpfreport
!!$
!!$  use sqpfmean_module
!!$  use iso_fortran_env
!!$  implicit none
!!$
!!$  type, extends(sqpfmean_diagnose) :: sqpfmean_diagnosis
!!$!     integer :: unit
!!$   contains
!!$     procedure :: report => stepprint
!!$  end type sqpfmean_diagnosis
!!$
!!$contains
!!$
!!$  subroutine stepprint(this,iter,t,s,f,dlt,p,g,B,iflag)
!!$    class(sqpfmean_diagnosis) :: this
!!$    integer, intent(in) :: iter, iflag
!!$    real(REAL64), intent(in) :: t,s,f,dlt
!!$    real(REAL64), dimension(2), intent(in) :: p,g
!!$    real(REAL64), dimension(2,2), intent(in) :: B
!!$
!!$!    write(this%unit,*) iter,t,s,f,dlt,p,g,B
!!$
!!$  end subroutine stepprint
!!$
!!$end module sqpfreport

program testsr1mean_hist

  ! test the estimator

  use toolbox
  use rfun_REAL64
  use rfun_hampel
!  use sr1mean_function
  use rlike_REAL64
  use iso_fortran_env
  implicit none

  character(len=15), parameter :: robfun = 'tukey'
  integer, parameter :: datamax = 5
  integer, parameter :: ndat = 1000
  integer, parameter :: k = datamax
  real(REAL64), parameter :: pi = 3.141592653589793_REAL64
  real(REAL64), parameter :: tpar =  0e0
  real(REAL64), parameter :: sigpar = 1e0
  real(REAL64), parameter :: beta = 0.8718
  real(REAL64), parameter :: s0 = sqrt(beta)
  real(REAL64), parameter :: gama = 0.5956
  real(REAL64), parameter :: nu = beta
  real(REAL64), parameter :: hnorm = 2.57657
  integer, parameter :: step = 0
  real(REAL64), parameter :: reltol = 1.0 / 42.0 * 1e-5
!  real(REAL64), dimension(:), pointer :: datas
  real(REAL64), dimension(:), allocatable :: t,s,sig, med, Qn, data, chi2, sig2, &
       res, avgR
  real(REAL64), dimension(:), allocatable, target :: dataset
  real(REAL64), dimension(2,2) :: hess
  real(REAL64) :: t0, t1, dt, d, mu, logmu, at,asig, fdat, hbin
  integer :: i, l, nfail, info, niter, nfev, nrst, nout, nidf, nhrd, nbin

  allocate(t(ndat),s(ndat),sig(ndat),dataset(datamax*ndat),med(ndat),Qn(ndat),&
       data(datamax),chi2(ndat),sig2(ndat),res(datamax),avgR(ndat))

  nfail = 0
  niter = 0
  nfev = 0
  nrst = 0
  nout = 0
  nidf = 0
  nhrd = 0
  hess = 0
  do i = 1, ndat
     call rmean_new(t(i),s(i),sig(i),med(i),Qn(i),data,info)
     l = i*datamax
     dataset(1+(l-datamax):l) = data
     at = sum(data)/datamax
     asig = sqrt(sum((data-at)**2)/(datamax-1))
     res = (data-t(i)) / s(i)
     chi2(i) = sig(i)**2*(datamax-1)
     avgR(i) = sum(ipshampel(res)) / size(res)
!     chi2(i) = sig(i)**2*(datamax-1.0)/datamax
!     chi2(i) = 1 / chi2(i)
!     chi2(i) = s(i)**2 * sum(pshampel(res)**2) / sum(dpshampel(res))**2 *(datamax-1)**2
!     chi2(i) = sum(pshampel(res)**2) / sum(dpshampel(res))**2 * datamax**2
!     chi2(i) = sum(pshampel(res)**2)
!     chi2(i) = s0**2 *sum(pshampel(res)**2) / sum(dpshampel(res))**2 *datamax**2
!     chi2(i) = s(i)**2 * sum(ipshampel(res)) / sum(dpshampel(res))**2 *datamax**2
!     chi2(i) = 2*sum(ipshampel(res))*s(i)**2/sig(i)
!     chi2(i) = 2*sum(ipshampel((data-t(i))/sig(i)))
!     chi2(i) = 2*sum(ipshampel(data))
!     chi2(i) = sum(pshampel(data)**2)
!     chi2(i) = (s(i)/sig(i))**2 * sum(res**2)
!     chi2(i) = sum(data**2)
!     chi2(i) = sum((data-at)**2/asig)
!     chi2(i) = sum(data**2)
     if( .not. (info == 1 .or. info == 2) ) nfail = nfail + 1
!     sig(i) = sqrt(chi2(i))
  end do
!  sig2 = sig**2*datamax

  hess = hess / ndat
  write(*,'("mean H: ",2(2x,g0))') hess
  write(*,*) 'mean t:',sum(t)/ndat
  write(*,*) 'mean s:',sum(s)/ndat
  write(*,*) 'mean stdsig:',sum(sig)/ndat

  write(*,*) 'Convergence failed in ',nfail,' case(s).'
  write(*,*) 'Total niter=',niter,' nfev=',nfev,', nres=',nrst, ' nout=',nout, &
       ' nidf=',nidf,' nhrd=',nhrd
  fdat = ndat
  write(*,'(6(a,f0.3))') &
       'Average  niter=',niter/fdat,' nfev=',nfev/fdat,', nres=',nrst/fdat,&
       ' nout=',nout/fdat,' nidf=',nidf/fdat,' nhrd=',nhrd/fdat

!  stop

  write(error_unit,'(a)',advance="no") "Calculating graphs .. "

  d = sigpar/sqrt(1.0*datamax)
  t0 = tpar - 5*d
  t1 = tpar + 5*d
  dt = d/20
  hbin = 3.5*d/ndat**(1.0/3.0)
  nbin = (t1 - t0) / hbin
  call histogram(t,t0,t1,nbin,'/tmp/testsr1mean_t',.true.)
  call density_loc(t0,t1,dt,datamax,'/tmp/testsr1mean_density_t')
  d = sigpar/sqrt(2.0*datamax)
  t0 = max(sigpar - 5*d,0.0_REAL64)  ! 0.45
  t1 = sigpar + 5*d  ! 1.45
  dt = d/20
  call histogram(s,t0,t1,nbin,'/tmp/testsr1mean_s',.true.)
  call density_s(t0,t1,dt,datamax,'/tmp/testsr1mean_density_s')
  logmu = log(2.0)/2 + log_gamma((k+1)/2.0) - log_gamma(k/2.0)
  mu = exp(logmu)
  d = sqrt(k - mu**2)
  t0 = max(mu - 5*d,0.0_REAL64)  !3
  t1 = mu + 5*d  !10
  dt = d/20
  !  call histogram(sig*sqrt(1.0*datamax)/sigpar,t0,t1,150,'/tmp/testsr1mean_sig',.true.)
  d = sigpar/sqrt(datamax-2.0)
  t0 = max(sqrt(2.0)*sigpar - 5*d,0.0_REAL64)  ! 0.45
!  t0 = sigpar - 5*d  ! 0.45
  t1 = sigpar + 5*d  ! 1.45
!  t0 = 3
!  t1 = 11
!  call histogram(sig*sqrt(datamax-1.0),t0,t1,150,'/tmp/testsr1mean_sig',.true.)
!  call density_sig(t0,t1,dt,datamax,'/tmp/testsr1mean_density_sig')

  t0 = 0.45
  t1 = 0.75
  call histogram(avgR,t0,t1,nbin,'/tmp/testsr1mean_avgR',.true.)

  d = sqrt(2.0*datamax)
  t0 = max(k - 5*d,0.0_REAL64)
  t1 = k + 5*d
  dt = d/20
  call histogram(chi2/sigpar**2,t0,t1,nbin,'/tmp/testsr1mean_chi2',.true.)
  call density_chi2(t0,t1,dt,datamax,'/tmp/testsr1mean_density_chi2')



  block
    real(REAL64), parameter :: tmin =-3e0
    real(REAL64), parameter :: tmax = 3e0
    real(REAL64), parameter :: smin = 0e0
    real(REAL64), parameter :: smax = 3e0
    type(robust_likelihood_hampel), target :: hampelfun
    type(robust_likelihood_tukey), target :: tukeyfun
    type(robust_likelihood_huber), target :: huberfun
    type(nonrobust_likelihood_square), target :: squarefun
    class(robust_likelihood_REAL64), pointer :: fcn


    if( robfun == 'huber' ) then
       fcn => huberfun
    else if( robfun == 'tukey' ) then
       fcn => tukeyfun
    else if( robfun == 'square' ) then
       fcn => squarefun
    else !if( robfun == 'hampel' ) then
       fcn => hampelfun
    end if
    fcn%data => dataset

    call rland(fcn,tmin,tmax,smin,smax,'/tmp/testsr1mean_hist_rland')
    !    call chi2land(fcn,tmin,tmax,smin,smax,'/tmp/testsr1mean_hist_rland')

    open(1,file='/tmp/testsr1mean_hist_rland.par')
    write(1,'(a,g0)') 'TMIN = ',tmin
    write(1,'(a,g0)') 'TMAX = ',tmax
    write(1,'(a,g0)') 'SMIN = ',smin
    write(1,'(a,g0)') 'SMAX = ',smax
    close(1)

  end block

  open(1,file='/tmp/testsr1mean_hist')
  do i = 1, ndat
     write(1,*) t(i),s(i),sig(i),chi2(i)
  end do
  close(1)

  open(1,file='/tmp/testsr1mean_hist.par')
  write(1,'(i0)') datamax
  close(1)

  open(1,file='/tmp/testsr1mean_hist_rank')
  do i = 1, ndat
     write(1,*) med(i),Qn(i)
  end do
  close(1)

  open(1,file='/tmp/testsr1mean_hist_data')
  write(1,'(a)') '# 6*N(0,1) + 1*N(-10,1), Hampel'
  write(1,*) datamax*ndat
  write(1,*) dataset
  close(1)

  write(error_unit,*) "done"


contains

  subroutine rmean_new(t,s,sig,med,Qn,data,info)

!    use trustsr1mean
    use noise_REAL64
!    use quantilmean

    real(REAL64), intent(out) :: t,s,sig,med,Qn
    real(REAL64), dimension(:), intent(out), target :: data
    integer, intent(out) :: info

!    character(len=80) :: msg
!    real(REAL64) :: stderr, stddev
    integer :: i, istep

    call gnoise_fill(tpar,sigpar,data)
    if( step > 0 ) then
       istep = step
       do i = 2, datamax, istep
          data(i) = gnoise(-.0_REAL64,100.0_REAL64)
          !       write(*,*) data(i)
       end do
    end if

    ! save data
    open(3,file='/tmp/tsr1mean_data',form='unformatted')
    write(3) datamax
    write(3) data
    close(3)

!    goto 33
!    call qmean(data,med,stdsig=Qn)
!    t = qmed(data)
!    s = byQ(data)
!    med = t
!    Qn = s
!    Qn = qmed(abs(data-t)) / 0.6745
!    s = Qn
!    s = 0.8*s
!    call trmean(data,t,s,stderr,stddev,robfun,info,msg,.true.)
!    sig = stddev!*sqrt(real(datamax,REAL64))
!    sig = sqrt(sum(data**2)/(size(data)))
!    sig = sqrt(sum((data-t)**2))
!    sig = sum((data-t)**2)
    !    t = sum(data)/datamax

!    if( info /= 0 ) then
!       write(*,*) info, msg
!       stop
!    end if

!    if( s < 0.1 ) stop 's < 0.1'

!    if( info /= 0 ) stop 'trmean() failed.'
!33  continue

    block
      use robustmean_REAL64
!      use sqpfreport
      !      type(sqpfmean_diagnosis) :: diagnose
      character(len=*), parameter :: report_file = "/tmp/tsr1mean_hist_report"
      real(REAL64) :: mean,stderr,stdsig,scale
      real(REAL64), dimension(2,2) :: mhess
      integer :: iflag,miter,mfev,mrst,mout,midf,mhrd,eq,iostat
      character(len=666) :: line
      mean = tpar
      scale = 0.9*sigpar
      mean = med
      scale = Qn
      mean = sum(data)/size(data)
      scale = sqrt(sum((data-mean)**2)/(size(data)))
      call rmean_REAL64(data,mean,stderr,stdsig,scale,reltol,robfun,.true., &
           verbose=.true.,flag=iflag,report=report_file)

      open(13,file=report_file,status='old')
      do
         read(13,'(a)',iostat=iostat) line
         if( iostat == IOSTAT_END ) exit
         eq = index(line,'=') + 1
         if( index(line,'t0') > 0 ) read(line(eq:),*) med
         if( index(line,'s0') > 0 ) read(line(eq:),*) Qn
         if( index(line,'niter') > 0 ) read(line(eq:),*) miter
         if( index(line,'nfev') > 0 ) read(line(eq:),*) mfev
         if( index(line,'nrst') > 0 ) read(line(eq:),*) mrst
         if( index(line,'nout') > 0 ) read(line(eq:),*) mout
         if( index(line,'nidf') > 0 ) read(line(eq:),*) midf
         if( index(line,'nhrd') > 0 ) read(line(eq:),*) mhrd
         if( index(line,'hess') > 0 ) read(line(eq:),*) mhess
      end do
      close(13)
!           verbose=.true.,diagnose=diagnose,flag=iflag)
!      write(error_unit,'(a,g0,3(2x,es10.2),3(a,i0))') &
!           'rmean:',mean,scale,stdsig,stderr, &
!           '  nfev=',diagnose%nfev,' nrst=',diagnose%nrst,' iflag=',iflag
!      write(*,*) 'rmean diagnose: ','nfev=',diagnose%nfev,' nrestores=',diagnose%nrst
      t = mean
      s = scale
      sig = stdsig
      info = iflag
      hess = hess + mhess
      niter = niter + miter
      nfev = nfev + mfev
      nrst = nrst + mrst
      nout = nout + mout
      nidf = nidf + midf
      nhrd = nhrd + mhrd
!      hess = hess + diagnose%hess
!      niter = niter + diagnose%niter
!      nfev = nfev + diagnose%nfev
!      nrst = nrst + diagnose%nrst
!      nout = nout + diagnose%nout
!      nidf = nidf + diagnose%nidf
!      nhrd = nhrd + diagnose%nhrd
!      med = diagnose%t0
!      Qn = diagnose%s0
      if( info >= 5 ) stop 'info >= 5'
!      if( diagnose%nrst > 0 ) stop 'nrest > 0'
    end block

  end subroutine rmean_new


!!$  subroutine rmean(t,s,sig)
!!$
!!$    use trustsr1mean
!!$!    use symmetric_rank_1
!!$    use noise
!!$    use quantilmean
!!$
!!$    real(REAL64), intent(out) :: t,s,sig
!!$    !    type(sr1_mean) :: funmean
!!$    type(tsr1mean_tukey), target :: tukeymean
!!$    type(tsr1mean_huber), target :: hubermean
!!$!    type(tsr1mean_square), target :: squaremean
!!$    type(tsr1mean_experimental), target :: bathmean
!!$    class(tsr1mean_function), pointer :: funmean
!!$
!!$    real(REAL64), dimension(3), parameter :: q = [0.25, 0.5, 0.75]
!!$!    real(REAL64), dimension(3), parameter :: qq = [0.4, 0.6745, 0.87]
!!$    real(REAL64), dimension(3), parameter :: qq = [0.4, 0.5, 0.8]
!!$    real(REAL64), dimension(size(q)) :: f,g
!!$    real(REAL64), dimension(:), allocatable, target :: data
!!$!    real(REAL64), dimension(2) :: x
!!$    real(REAL64), dimension(2,2) :: cov
!!$    real(REAL64) :: delta, maxdev, dt, smin
!!$    integer :: maxfev, info, flag, i, iter
!!$    character(len=80) :: msg
!!$    logical :: verbose, quasi
!!$
!!$    allocate(data(datamax))
!!$    call gnoise_fill(0.0_REAL64,1.0_REAL64,data)
!!$    do i = 2, 0*datamax, 26
!!$       data(i) = gnoise(-25.0_REAL64,1.0_REAL64)
!!$!       write(*,*) data(i)
!!$    end do
!!$
!!$    funmean => bathmean
!!$!    funmean => tukeymean
!!$!    funmean => hubermean
!!$    funmean%data => data
!!$
!!$    call quantilefun(data,q,f,flag)
!!$!    t = (f(2) + (f(3) + f(1))/2)/2
!!$    t = f(2)
!!$
!!$    call quantilefun(abs(data-t),qq,g,flag)
!!$!    s = (f(3) - f(1)) / (2*0.6745)
!!$    s = g(2) / 0.6745
!!$
!!$!    s = (f(3) - f(1)) / (2*0.6745)
!!$
!!$!    call quantilefun(abs(data-t),qq,g,info)
!!$!    s = g(2)
!!$
!!$    write(*,*) 'init:',real(t),real(s)
!!$
!!$  block
!!$    use fmm
!!$    use likescale
!!$
!!$    real(REAL64) :: smax,tol,rmin,rmax
!!$    logical :: found
!!$
!!$    datas => data
!!$
!!$!    tol = 0.01*s
!!$!    dft = t
!!$!    rmin = sqrt(epsilon(rmin))
!!$!    rmax = s
!!$!    smin = zeroin(rmin,rmax,scalemin,tol)
!!$
!!$    ! Does the minimum falls inside the interval smin < .. < smax ?
!!$!    found = rmin + tol < smin .and. smin < rmax - tol
!!$
!!$    !    if( debug )
!!$!    write(*,'(a,g20.6,a,g0.2,l2,i5)') &
!!$!         'Minimal scale limit:',smin," +- ",tol,found
!!$
!!$    write(*,*) g
!!$    tol = 0.1*g(2)
!!$!    tol = 0.1*s
!!$    !    smin = g(1)
!!$    smin = g(1)
!!$    smax = g(3)
!!$!    smax = maxval(abs(data-t))
!!$    tol = (smax - smin) / 1000
!!$    chi2t = t
!!$    write(*,*) '*',real(smin),real(smax),real(tol)
!!$!    s = fmin(5*smin,smax,chi2min,tol)
!!$!    s = zeroin(smin,smax,chi2fun,tol)
!!$
!!$    ! Does the minimum falls inside the interval smin < .. < smax ?
!!$    found = smin + tol < s .and. s < smax - tol
!!$
!!$    call iscale(abs(data - t),smin,verbose=.true.)
!!$    write(*,*) 'by iscale:',smin
!!$!    s = smin
!!$
!!$    !    if( debug )
!!$    write(*,'(a,g20.6,a,g0.2,l2,i5)') &
!!$         'Scale initial estimation:',s," +- ",tol,found
!!$
!!$    datas => NULL()
!!$  end block
!!$
!!$  block
!!$    use kstestscale
!!$
!!$!    s = kscale(data-t)
!!$    write(*,*) 'by K-S test optim:',s
!!$
!!$  end block
!!$
!!$!  block
!!$!    logical :: reli
!!$!    integer :: iflag
!!$
!!$!    call qmean(data,t,dt,s,.false.,reli,iflag)
!!$!    s = (s - smin) / 2 + smin
!!$!  end block
!!$
!!$
!!$!    x = [ t, s ]
!!$    delta = 0.1 * s !/ sqrt(real(datamax))
!!$    maxdev = 1e-2*s/sqrt(real(datamax))
!!$    maxfev = 3*precision(t)
!!$    verbose = .true.
!!$    quasi = .false.
!!$
!!$    do iter = 1,1
!!$ !   call sr1rob(funmean,x,cov,delta,maxdev,maxfev,quasi,verbose,info,msg)
!!$       !       call tsr1mean(data,t,dt,s,info,verbose)
!!$       block
!!$         logical :: reli
!!$         logical, parameter :: quasi = .true.
!!$         integer, parameter :: maxfev = 3*precision(t)
!!$         real(REAL64), dimension(2,2) :: hess
!!$         real(REAL64) :: tol,delta
!!$
!!$!         t = 0
!!$!         s = 1.5
!!$
!!$         tol = 0.01 * s / sqrt(real(size(data)))
!!$         delta = 0.1*s
!!$
!!$         call tsr1meansqp(funmean,t,s,hess,delta,tol,maxfev,quasi,verbose,info)
!!$         dt = funmean%stderr(t,s,reli)
!!$
!!$       end block
!!$
!!$    if( info /= 1 ) then
!!$!    if( s < 1e-2 ) then
!!$       write(*,*) 'data:',data
!!$       write(*,*) 'solution:',real(t),real(s),real(dt)
!!$
!!$       call graph(funmean,-5.0,5.0,0.0,5.0,'/tmp/sr1mean_hist')
!!$       !       call graphchi2(funmean,-5.0,5.0,0.0,5.0,'/tmp/sr1mean_stddev_hist')
!!$       call graphi2(funmean,data,t,0.0,5.0,'/tmp/sr1mean_chi2')
!!$
!!$       ! save data
!!$       open(3,file='/tmp/tsr1mean_data',form='unformatted')
!!$       write(3) datamax
!!$       write(3) data
!!$       close(3)
!!$
!!$       if( iter == 2 ) then
!!$          if( info == 5 ) nfail = nfail + 1
!!$!          if( info == 5 ) stop 'info == 5'
!!$       else
!!$          if( info == 5 ) then
!!$             s = 2*g(2) / 0.6745
!!$             t = f(2)
!!$          end if
!!$       end if
!!$       stop 'info /= 0'
!!$    end if
!!$ end do
!!$!    t = x(1)
!!$!    s = x(2)
!!$!    sig = funmean%stderr(t,s)*sqrt(real(datamax))
!!$    sig = dt*sqrt(real(datamax))
!!$
!!$!    block
!!$!      t = sum(data)/datamax
!!$!      sig = sqrt(sum((data-t)**2)/(datamax-1))
!!$!    end block
!!$
!!$
!!$  end subroutine rmean
!!$
!!$  subroutine graph(fcn,x0,x1,y0,y1,filename)
!!$
!!$    class(tsr1mean_function), intent(in) :: fcn
!!$!    class(sr1_function), intent(in) :: fcn
!!$    real(REAL32), intent(in) :: x0,x1,y0,y1
!!$    character(len=*), intent(in) :: filename
!!$    integer, parameter :: nx = 200
!!$    integer, parameter :: ny = 200
!!$    real(REAL64) :: x,dx,y,dy
!!$    integer :: i,j
!!$
!!$    dx = (x1 - x0) / (nx - 1)
!!$    dy = (y1 - y0) / (ny - 1)
!!$
!!$    open(1,file=filename)
!!$    do j = 1, ny
!!$       y = y0 + dy*j
!!$       do i = 0, nx
!!$          x = x0 + dx*i
!!$          write(1,*) x,y,fcn%val(x,y)
!!$       end do
!!$       write(1,*)
!!$    end do
!!$    close(1)
!!$
!!$  end subroutine graph
!!$
!!$  subroutine graphi2(fcn,data,t,s0,s1,filename)
!!$
!!$    use rfun
!!$
!!$    class(tsr1mean_function), intent(in) :: fcn
!!$    real(REAL64), dimension(:), intent(in) :: data
!!$    real(REAL64), intent(in) :: t
!!$    real(REAL32), intent(in) :: s0,s1
!!$    character(len=*), intent(in) :: filename
!!$    integer, parameter :: n = 200
!!$    real(REAL64) :: s,ds,chi2
!!$    integer :: i
!!$
!!$    ds = (s1 - s0) / (n - 1)
!!$
!!$    open(1,file=filename)
!!$    do i = 1, n
!!$       s = s0 + ds*i
!!$       chi2 = sum(tukey((data - t)/s)**2) / size(data) - stukey
!!$       write(1,*) s,chi2,sum(pshuber((data - t)/s)**2) / size(data) -spshuber, &
!!$            sum(((data - t)/s)**2) / size(data) - 1
!!$    end do
!!$    close(1)
!!$
!!$  end subroutine graphi2


!!$  subroutine graphdev(fcn,x0,x1,y0,y1,filename)
!!$
!!$    class(sr1_mean), intent(in) :: fcn
!!$    real(REAL32), intent(in) :: x0,x1,y0,y1
!!$    character(len=*), intent(in) :: filename
!!$    integer, parameter :: nx = 200
!!$    integer, parameter :: ny = 200
!!$    real(REAL64) :: x,dx,y,dy,d
!!$    integer :: i,j
!!$
!!$    dx = (x1 - x0) / (nx - 1)
!!$    dy = (y1 - y0) / (ny - 1)
!!$
!!$    open(1,file=filename)
!!$    open(2,file='/tmp/c')
!!$    do j = 1, ny
!!$       y = y0 + dy*j
!!$       do i = 1, nx
!!$          x = x0 + dx*i
!!$          d = fcn%stderr([x,y])
!!$          write(1,*) x,y,d
!!$          if( abs(d - 0.9) < 0.01 ) write(2,*) x,y
!!$       end do
!!$       write(1,*)
!!$    end do
!!$    close(1)
!!$
!!$    close(2)
!!$
!!$  end subroutine graphdev

!!$  real(REAL64) function scalemin(s)
!!$    use rfun
!!$    real(REAL64), intent(in) :: s
!!$    real(REAL64) :: t
!!$    t = dft
!!$    if( .not. associated(datas) ) stop 'not associated'
!!$    scalemin = sum(dtukey((datas-t)/s), dtukey((datas-t)/s) > 0) / size(datas) - 0.3
!!$  end function scalemin

!!$  real(REAL64) function chi2min(s)
!!$    use rfun
!!$    real(REAL64), intent(in) :: s
!!$    chi2min = sum(itukey((datas-chi2t)/s))/size(datas) + log(s)
!!$  end function chi2min
!!$
!!$  real(REAL64) function chi2fun(s)
!!$    use rfun
!!$    real(REAL64), intent(in) :: s
!!$!    chi2fun = sum(pshuber((datas-chi2t)/s)**2)/(size(datas)-1) - spshuber
!!$    chi2fun = sum(pshampel((datas-chi2t)/s)**2)/(size(datas)-1) - spshuber
!!$  end function chi2fun
!!$

  subroutine rland(fcn,x0,x1,y0,y1,filename)

    class(robust_likelihood_REAL64), intent(in) :: fcn
    real(REAL64), intent(in) :: x0,x1,y0,y1
    character(len=*), intent(in) :: filename
    integer, parameter :: nx = 200
    integer, parameter :: ny = 200

    real(REAL64) :: x,dx,y,dy
    integer :: i,j

    dx = (x1 - x0) / (nx - 1)
    dy = (y1 - y0) / (ny - 1)

    open(1,file=filename)
    do j = 1, ny
       y = y0 + dy*j
       do i = 0, nx
          x = x0 + dx*i
          write(1,*) x,y,fcn%func(x,y)
       end do
       write(1,*)
    end do
    close(1)

  end subroutine rland

!!$  subroutine chi2land(fcn,x0,x1,y0,y1,filename)
!!$
!!$    class(robust_likelihood), intent(in) :: fcn
!!$    real(REAL64), intent(in) :: x0,x1,y0,y1
!!$    character(len=*), intent(in) :: filename
!!$    integer, parameter :: nx = 200
!!$    integer, parameter :: ny = 200
!!$
!!$    real(REAL64) :: x,dx,y,dy,fval
!!$    integer :: i,j
!!$
!!$    dx = (x1 - x0) / (nx - 1)
!!$    dy = (y1 - y0) / (ny - 1)
!!$
!!$    open(1,file=filename)
!!$    do j = 1, ny
!!$       y = y0 + dy*j
!!$       do i = 0, nx
!!$          x = x0 + dx*i
!!$          fval = fcn%func(x,y) - log(y)
!!$          write(1,*) x,y,fval!sum(hampel((dataset - x)/y)**2)
!!$       end do
!!$       write(1,*)
!!$    end do
!!$    close(1)
!!$
!!$  end subroutine chi2land


  subroutine density_loc(t0,t1,dt,N,somefile)

    real(REAL64), intent(in) :: t0,t1,dt
    integer, intent(in) :: N
    character(len=*), intent(in) :: somefile
    real(REAL64) :: t,sig,f
    integer :: i

    sig = sigpar / sqrt(real(N))
    open(1,file=somefile)
    do i = 1, int((t1 - t0) / dt)
       t = t0 + i*dt
       f = exp(-(t - tpar)**2/2/sig**2) / (sqrt(2*pi) * sig)
       write(1,*) t,f
    end do
    close(1)

  end subroutine density_loc

  subroutine density_s(s0,s1,ds,N,somefile)

    real(REAL64), intent(in) :: s0,s1,ds
    integer, intent(in) :: N
    character(len=*), intent(in) :: somefile
    real(REAL64) :: s,sig,f,f1,t,sig0
    integer :: i

    t = 0
!    sig = sigpar* beta**2 / sqrt(2.0*real(N))
!    sig = s0 / sqrt((1 + gama)*N)
    sig = sigpar * sqrt(beta/((1 + gama)*N))
    sig0 = sqrt(beta)
    open(1,file=somefile)
    do i = 1, int((s1 - s0) / ds)
       s = s0 + i*ds
!       f1 = exp(-(t - sig0*sigpar)**2/2/sig**2) / (sqrt(2*pi) * sig)
       f1 = exp(-(s - sig0*sigpar)**2/2/sig**2) / (sqrt(2*pi) * sig)
!       f = exp(-ipshampel((t - sig0*sigpar)/sig)) / (hnorm * sig)
!       f = exp(-ipshampel((s - sig0*sigpar)/sig)) / (hnorm * sig)
       write(1,*) s,f,f1
    end do
    close(1)

  end subroutine density_s

  subroutine density_sig(t0,t1,dt,N,somefile)

    real(REAL64), intent(in) :: t0,t1,dt
    integer, intent(in) :: N
    character(len=*), intent(in) :: somefile
!    integer, parameter :: m = 10000
    real(REAL64), dimension(:), allocatable :: t,f,g,h
    real(REAL64) :: s,sig,logk,u
    integer :: i, j, k, m

    !    m = int((t1 - t0) / dt)
    m = int(t1/dt)
    allocate(t(m),f(m),g(-m:m),h(m))

    k = N !- 1
    logk = (k/2.0 - 1)*log(2.0) + log_gamma(k/2.0)
    do i = 1, size(h)
       t(i) = 0*t0 + i*dt
       u = (k-1)*log(t(i)) - t(i)**2/2 - logk
       h(i) = exp(u)
    end do
    h = h * dt

    sig = sigpar*sqrt(beta/((1+gama)*N))
!    sig = 1.071/sqrt(2.0*k)
    g(0) = 1 / sqrt(2*pi)/sig
    do i = 1, m
       g(i) = exp(-(t(i)-0)**2/2/sig**2)/sqrt(2*pi)/sig
       g(-i) = g(i)
    end do
    g = g * dt

    ! the convolution
    do i = 1, size(f)
       s = 0
       do j = -2*m, 2*m
          if( 1 <= j .and. j <= m .and. -m <= i-j .and. i-j <= m ) then
             s = s + h(j)*g(i-j)
          end if
       end do
       f(i) = s / dt
    end do
    f = g / dt

    open(1,file=somefile)
    do i = int(t0/dt+1), size(t)
       write(1,*) t(i),f(i)
    end do
    close(1)

  end subroutine density_sig


  subroutine density_chi2(t0,t1,dt,N,somefile)

    real(REAL64), intent(in) :: t0,t1,dt
    integer, intent(in) :: N
    character(len=*), intent(in) :: somefile
    real(REAL64), dimension(:), allocatable :: t,f,g,h
    real(REAL64) :: s,sig,logk2,u2
    integer :: i, j, k, m

!    m = int((t1 - t0) / dt)
    m = int(t1/dt)
    allocate(t(m),f(m),g(-m:m),h(m))

    k = N !- 1
    logk2 = (k/2.0) * log(2.0) + log_gamma(k/2.0)    ! chi2
    do i = 1, size(h)
       t(i) = 0*t0 + i*dt
       u2 = (k/2.0-1.0)*log(t(i)) - t(i)/2 - logk2   ! chi2
       h(i) = exp(u2)
    end do
    h = h *dt

!    sig = beta/sqrt(2*real(N))
    sig = sigpar*sqrt(beta/((1+gama)*N))
!    sig = 1.071/sqrt(2.0*k)
    g(0) = 1 / sqrt(2*pi)/sig
    do i = 1, m
       g(i) = exp(-(t(i)-0)**2/2/sig**2)/sqrt(2*pi)/sig
       g(-i) = g(i)
    end do
    g = g * dt

    ! the convolution
    do i = 1, size(f)
       s = 0
       do j = -2*m, 2*m
          if( 1 <= j .and. j <= m .and. -m <= i-j .and. i-j <= m ) then
             s = s + h(j)*g(i-j)
          end if
       end do
       f(i) = s / dt
    end do

    open(1,file=somefile)
    do i = int(t0/dt+1), size(t)
       write(1,*) t(i),f(i)
    end do
    close(1)

  end subroutine density_chi2

  real(REAL64) function stirlig(x)

    real(REAL64), intent(in) :: x
    real(REAL64), parameter :: pi = 3.141592653589793_REAL64
    real(REAL64), parameter :: e = 2.718281828459045_REAL64

    stirlig = sqrt(2*pi*x)*(x/e)**x

  end function stirlig


end program testsr1mean_hist
